/***********************************************************************
 HAF : Heuristic Alignment with Frameshifts
 bmajoros@tigr.org

 BOOM : Bioinformatics Object Oriented Modules

 This software is open source, and is distributed under the terms of the
 ARTISTIC LICENSE, as described at www.opensource.org.
 ***********************************************************************/
#include "HAF.H"
#include <iostream>
#include <math.h>
#include "AminoAlphabet.H"
#include "Constants.H"
#include "Stack.H"
using namespace std;
using namespace BOOM;


HAF::HAF(const Sequence &S1,const SubstitutionMatrix<float> &M,
	 const Sequence &S2_frame0,const Sequence &S2_frame1,
	 const Sequence &S2_frame2,
	 float gapPenalty,int bandWidth,float frameshiftPenalty)
  : gapPenalty(gapPenalty),
    frameshiftPenalty(frameshiftPenalty),
    bandWidth(bandWidth),
    M(M),
    alphabet(AminoAlphabet::global)
{
  // ctor

  this->S1=&S1;
  this->S2[0]=&S2_frame0;
  this->S2[1]=&S2_frame1;
  this->S2[2]=&S2_frame2;  
}



HAF::~HAF()
{
  // dtor

  delete alignmentMatrix;
}



Alignment *HAF::go(Sequence *&S2_composite)
{
  // Allocate alignment matrix
  xDim=S1->getLength()+1;
  yDim=max(max(S2[0]->getLength(),S2[1]->getLength()),S2[2]->getLength())+1;
  alignmentMatrix=new BandedAlignmentMatrix<Cell>(xDim,yDim,bandWidth);

  // Evaluate all cells in the matrix
  evaluateMatrix();

  // Track back to reconstruct the alignment
  return traceBack(S2_composite);

  // ### Also need to postprocess the alignment to coalesce into large
  //     frame-consistent regions, to supress over-aggressive frame shifting
}



void HAF::evaluateMatrix()
{
  // Initialize the bottom row and leftmost column of the dynamic 
  // programming matrix
  Cell *origin;
  if(!alignmentMatrix->getCell(0,0,origin)) throw "HAF 1";
  for(int f=0 ; f<3 ; ++f) origin->frame[f].score=0;
  for(int i=1 ; i<xDim ; ++i)
    {
      Cell *bottomCell, *pred=origin;
      if(!alignmentMatrix->getCell(i,0,bottomCell)) break;
      for(int f=0 ; f<3 ; ++f) 
	{
	  Cell::Frame &cell=bottomCell->frame[f];
	  cell.link.set(LEFT,f);
	  cell.score=pred->frame[f].score+gapPenalty;
	}
      pred=bottomCell;
    }
  for(int i=1 ; i<yDim ; ++i)
    {
      Cell *leftCell, *pred=origin;
      if(!alignmentMatrix->getCell(0,i,leftCell)) break;
      for(int f=0 ; f<3 ; ++f)
	{
	  Cell::Frame &cell=leftCell->frame[f];
	  cell.link.set(DOWN,f);
	  cell.score=pred->frame[f].score+gapPenalty;
	}
      pred=leftCell;
    }

  // Fill out the main portion of the matrix
  for(int x=1 ; x<xDim ; ++x)
    for(int y=1 ; y<yDim ; ++y)
      {
	Cell *thisCell=NULL;
	if(!alignmentMatrix->getCell(x,y,thisCell)) continue;
	Cell *leftCell=NULL, *downCell=NULL, *diagonalCell=NULL;
	alignmentMatrix->getCell(x-1,y,leftCell);
	alignmentMatrix->getCell(x,y-1,downCell);
	alignmentMatrix->getCell(x-1,y-1,diagonalCell);
	evaluateCell(thisCell,leftCell,downCell,diagonalCell,x,y);
      }
}



void HAF::evaluateCell(HAF::Cell *thisCell,HAF::Cell *leftCell,
		       HAF::Cell *downCell,HAF::Cell *diagonalCell,
		       int x,int y)
{
  // Initialize to -INF
  for(int i=0 ; i<3 ; ++i) thisCell->frame[i].score=NEGATIVE_INFINITY;

  // Consider linking to the left
  if(leftCell)
    for(int i=0 ; i<3 ; ++i)
      {
	Cell::Frame &frame=thisCell->frame[i];
	for(int j=0 ; j<3 ; ++j)
	  {
	    float F=(i==j ? 0 : frameshiftPenalty);
	    float leftScore=leftCell->frame[j].score+gapPenalty+F;
	    if(leftScore>frame.score) frame.set(leftScore,LEFT,j);
	  }
      }

  // Consider linking downward
  if(downCell)
    for(int i=0 ; i<3 ; ++i)
      {
	Cell::Frame &frame=thisCell->frame[i];
	for(int j=0 ; j<3 ; ++j)
	  {
	    float F=(i==j ? 0 : frameshiftPenalty);
	    float downScore=downCell->frame[j].score+gapPenalty+F;
	    if(downScore>frame.score) frame.set(downScore,DOWN,j);
	  }
      }

  // Consider linking diagonally (i.e., a match)
  if(diagonalCell)
    for(int i=0 ; i<3 ; ++i)
      {
	Cell::Frame &frame=thisCell->frame[i];
	for(int j=0 ; j<3 ; ++j)
	  {
	    float F=(i==j ? 0 : frameshiftPenalty);
	    if(y-1>=S2[j]->getLength()) 
	      {/*cout<<"skipping..."<<endl;###*/continue;}
	    Symbol a=(*S1)[x-1];
	    Symbol b=(*S2[j])[y-1];
	    float matchScore=M(a,b);
	    float diagScore=diagonalCell->frame[j].score+F+matchScore;
	    if(diagScore>frame.score) frame.set(diagScore,DIAGONAL,j);
	  }
      }
}



void HAF::followLink(const HAF::Cell::Frame::Link &link,
		     int &x,int &y,int &frame)
{
  switch(link.dir)
    {
    case LEFT:
      --x;
      break;
    case DOWN:
      --y;
      break;
    case DIAGONAL:
      --x;
      --y;
      break;
    }
  frame=link.frame;
}



Alignment *HAF::traceBack(Sequence *&S2_composite)
{
  // Select the highest-scoring frame for the terminal cell
  Cell *cell;
  float bestScore=NEGATIVE_INFINITY;
  int bestFrame;
  int x=xDim-1, y=-1;
  for(int f=0 ; f<3 ; ++f)
    {
      int yDim=S2[f]->getLength()+1;
      if(!alignmentMatrix->getCell(xDim-1,yDim-1,cell)) 
	throw "HAF traceback";
      Cell::Frame &frame=cell->frame[f];
      if(frame.score>bestScore)
	{
	  bestFrame=f;
	  bestScore=frame.score;
	  y=yDim-1;
	}
    }
  
  // Trace back from the terminal cell by following links
  Stack<Cell::Frame::Link> pathStack;
  int frame=bestFrame;
  while(x!=0 || y!=0)
    {
      alignmentMatrix->getCell(x,y,cell);
      Cell::Frame::Link &link=cell->frame[frame].link;
      pathStack.push(link);
      followLink(link,x,y,frame);
    }

  // Reverse the stack contents to produce a forward-oriented
  // alignment
  S2_composite=new Sequence;
  Alignment &alignment=*new Alignment(*S1,*S2_composite,alphabet,bestScore);
  const int len=pathStack.size();
  Vector<int> frames;
  for(int i=0 ; i<len ; ++i)
    {
      Cell::Frame::Link link=pathStack.pop();
      //cout<<link.frame<<" ";//###
      switch(link.dir)
	{
	case LEFT:
	  alignment+=FIRST_UNMATCHED;
	  break;
	case DOWN:
	  alignment+=SECOND_UNMATCHED;
	  frames.push_back(link.frame);
	  break;
	case DIAGONAL:
	  alignment+=MATCH;
	  frames.push_back(link.frame);
	  break;
	}
    }
  //cout<<endl;//###

  // Construct the composite version of S2 by observing selected frame
  // shifts in the alignment
  int framesLen=frames.size();
  for(int i=0 ; i<framesLen ; ++i)
    S2_composite->append((*S2[frames[i]])[i]);
  //cout<<"COMPOSITE=";S2_composite->printOn(cout,AminoAlphabet::global);cout<<endl;


  return &alignment;
}


